from abc import ABCMeta, abstractmethod
from collections.abc import Iterator
from typing import final

from ..exception import MarkupyError
from .view import View


class Component(View, metaclass=ABCMeta):
    @abstractmethod
    def render(self) -> View: ...

    @final
    def __iter__(self) -> Iterator[str]:
        node = self.render()
        if isinstance(node, View):  # type: ignore[unused-ignore]
            yield from node
        else:
            raise MarkupyError(
                f"{type(self).__name__}.render() must return an instance of markupy.View (can be Element, Fragment or Component)"
            )



from collections.abc import Iterable, Iterator
from copy import copy
from typing import Any, final

from markupsafe import escape
from typing_extensions import Self

from ..exception import MarkupyError
from .view import View


def iter_node(node: Any, *, safe: bool = False) -> Iterator[str | View]:
    if node is None or node == "" or isinstance(node, bool):
        return
    # View is Iterable
    elif isinstance(node, View):
        yield node
    elif isinstance(node, Iterable) and not isinstance(node, str):
        for child in node:  # type: ignore[unused-ignore]
            yield from iter_node(child, safe=safe)
    elif safe:
        yield str(node)
    else:
        yield str(escape(node))


class Fragment(View):
    __slots__ = ("_children", "_shared", "_safe")

    def __init__(self, *, safe: bool = False) -> None:
        self._children: list[str | View] | None = None
        self._shared: bool = True
        self._safe: bool = safe

    def __copy__(self) -> Self:
        return type(self)()

    def __repr__(self) -> str:
        return "<markupy.Fragment>"

    # Use subscriptable [] syntax to assign children
    def __getitem__(self, node: Any) -> Self:
        if self._children is not None:
            raise MarkupyError(f"Illegal attempt to redefine children of `{self!r}`")

        if new_children := list(iter_node(node, safe=self._safe)):
            instance = self._new_instance()
            instance._children = new_children
            return instance

        return self

    def __iter__(self) -> Iterator[str]:
        if self._children is not None:
            for node in self._children:
                if isinstance(node, View):
                    yield from node
                else:
                    yield node

    @final
    def _new_instance(self: Self) -> Self:
        # When imported, elements are loaded from a shared instance
        # Make sure we re-instantiate them on setting attributes/children
        # to avoid sharing attributes/children between multiple instances
        if self._shared:
            obj = copy(self)
            obj._shared = False
            return obj
        return self



from markupy import Component, Fragment, View, tag


class ComponentElement(Component):
    def __init__(self, id: str):
        self.id = id

    def render(self) -> View:
        return tag.Div(id=self.id)


class ComponentFragment(Component):
    def render(self) -> View:
        return Fragment[tag.Div, tag.Img]


class ComponentInComponent(Component):
    def render(self) -> View:
        return Fragment[tag.Input, ComponentElement("inside")]


class ComponentAsComponent(Component):
    def render(self) -> View:
        return ComponentElement("other")


def test_component_element() -> None:
    assert str(ComponentElement("component")) == """<div id="component"></div>"""


def test_component_fragment() -> None:
    assert str(ComponentFragment()) == """<div></div><img>"""


def test_component_in_component() -> None:
    assert str(ComponentInComponent()) == """<input><div id="inside"></div>"""


def test_component_as_component() -> None:
    assert str(ComponentAsComponent()) == """<div id="other"></div>"""



from markupy import Fragment
from markupy.tag import Div, I, P, Tr


def test_render_direct() -> None:
    assert str(Fragment["Hello ", None, I["World"]]) == """Hello <i>World</i>"""


def test_render_as_child() -> None:
    assert (
        str(P["Say: ", Fragment["Hello ", None, I["World"]], "!"])
        == """<p>Say: Hello <i>World</i>!</p>"""
    )


def test_render_nested() -> None:
    assert str(Fragment[Fragment["Hel", "lo "], "World"]) == """Hello World"""


def test_render_embedded() -> None:
    assert (
        str(P[Fragment["Good ", I["morning"]], " ", I["World"]])
        == """<p>Good <i>morning</i> <i>World</i></p>"""
    )


def test_safe() -> None:
    assert str(Fragment['>"']) == """&gt;&#34;"""


def test_iter() -> None:
    assert list(Fragment["Hello ", None, I["World"]]) == [
        "Hello ",
        "<i>",
        "World",
        "</i>",
    ]


def test_element() -> None:
    result = Fragment[Div["a"]]
    assert str(result) == "<div>a</div>"


def test_multiple_element() -> None:
    result = Fragment[Tr["a"], Tr["b"]]
    assert str(result) == "<tr>a</tr><tr>b</tr>"


def test_list() -> None:
    result = Fragment[[Tr["a"], Tr["b"]]]
    assert str(result) == "<tr>a</tr><tr>b</tr>"


def test_none() -> None:
    result = Fragment[None]
    assert str(result) == ""


def test_string() -> None:
    result = Fragment["hello!"]
    assert str(result) == "hello!"
